# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

from django.utils import six
from django.core.mail import send_mail
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext as _
from django.template.loader import render_to_string
from django.conf import settings
from django.core.urlresolvers import reverse

from datetime import datetime, timedelta
import logging
import collections

logger = logging.getLogger(__name__)

from .models import MailConfirmation


def get_users(to):
    if not isinstance(to, collections.Iterable):
        to = (to,)
    else:
        to = set(to)
    return filter(lambda x: isinstance(x, get_user_model()), to)


def generate_confirmation(toconfirm,
                          success_template='',
                          success_url='',
                          confirmed=False,
                          forusers=(),
                          *args, **kwargs):
    """
    Generates a confirmation email object
    """
    try:
        confirm = MailConfirmation(
            toconfirm_object=toconfirm,
            confirm_success_url=success_url,
            confirm_success_template=success_template,
            confirmed=confirmed)

        confirm.save()
        users = get_users(forusers)
        if users:
            confirm.forusers.add(*users)
    except Exception, e:
        logger.error('Could not generate confirmation for mail_confirmation: '+str(e)) # noqa
        raise
    return confirm



def send_confirmation(confirm, request_template=None,
                      subject=_('Confirmation mail'),
                      sender=None,
                      to=(),
                      confirmurlname='mail_confirmation:url',
                      context=None,
                      *args, **kwargs):
    """
    Send a confirmation mail
    """

    if sender is None:
        sender = settings.DEFAULT_FROM_EMAIL
    if isinstance(to, six.string_types):
        to = [to]
    try:
        scheme = settings.MAIL_CONFIRMATION_SCHEME
    except:
        scheme = 'https'
    try:
        from django.contrib.sites.models import Site
        domain = Site.objects.get_current().domain
    except:
        domain = settings.SITE_DOMAIN or ""

    url = reverse(confirmurlname,
                  kwargs={'confirmationid': confirm.confirmationid})
    request_template = request_template or confirm.toconfirm_object._meta.app_label + '/mail_request.html' # noqa
    basecontext = {'confirmation_url': url,
                   'SITE_SCHEME': scheme,
                   'SITE_DOMAIN': domain}
    if context:
        basecontext.update(context)
    message = render_to_string(request_template, basecontext)
    return send_mail(subject, message, sender,
                     to, fail_silently=False)


def generate_and_send(toconfirm,
                      request_template=None,
                      success_template='',
                      success_url='',
                      forusers=(),
                      mail_function=None,
                      *args,
                      **kwargs):
    """
    Do both actions above
    """
    confirm = generate_confirmation(toconfirm,
                                    success_template, success_url,
                                    forusers=forusers,
                                    *args, **kwargs)
    if confirm:
        if mail_function:
            return mail_function(confirm,
                                 request_template,
                                 forusers=forusers,
                                 *args, **kwargs)
        else:
            return send_confirmation(confirm,
                                     request_template,
                                     forusers=forusers,
                                     *args, **kwargs)
    return confirm


def clear_stale():
    """
    Clears stale requests not confirmed
    """
    try:
        delta = settings.MAIL_CONFIRMATION_STALE_PERIOD
    except:
        delta = 30
    if delta:
        last_month = datetime.today() - timedelta(days=30)
        MailConfirmation.objects.filter(date_created__lte=last_month).delete()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='MailConfirmation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('confirmationid', models.CharField(unique=True, max_length=255, verbose_name='Confirmation id')),
                ('confirmed', models.BooleanField(default=False, verbose_name='Confirmed')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('confirm_success_template', models.CharField(max_length=200, verbose_name='Template that shows the user a success message on confirmation', blank=True)),
                ('confirm_success_url', models.CharField(max_length=200, verbose_name='Redirect the user to this url when the url is confirmed', blank=True)),
                ('toconfirm_id', models.PositiveIntegerField()),
                ('toconfirm_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



# -*- coding: utf-8 -*-
import re

from django.core import mail
from django.test import TestCase
from django.template import Template
from django.core.urlresolvers import reverse
from django.conf import settings

from ..utils import generate_and_send, generate_confirmation
from ..models import MailConfirmation
from .models import Confirmed

class MailConfirmationTest(TestCase):
    r = re.compile('(' + settings.SITE_SCHEME + '://' + settings.SITE_DOMAIN + '/.*)')

    def test_confirmed(self):
        obj = Confirmed()
        obj.save()
        confirm = generate_confirmation(obj, confirmed=True)
        self.assertEqual(confirm.confirmed, True)

    def test_defaults(self):
        obj = Confirmed()
        obj.save()
        t = 'confirm.html'
        ct = 'confirmed.html'

        ret = generate_and_send(obj, to='user@example.com')
        self.assertNotIn(ret, [None])
        self.assertEqual(len(mail.outbox)>=1, True)
        confirm_url = self.r.findall(mail.outbox[-1].body)[0]
        email = mail.outbox[-1]
        self.assertEqual(email.to[0], 'user@example.com')
        self.assertEqual(email.body, u'confirm ' + settings.SITE_SCHEME + '://' + settings.SITE_DOMAIN + reverse('mail_confirmation:url', kwargs={'confirmationid': obj.confirmed.all()[0].confirmationid})+ '\n')
        self.assertEqual(email.from_email, 'webmaster@localhost')
        self.assertEqual(email.subject, 'Confirmation mail')
        response = self.client.get(confirm_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.template_name[0], 'tests/mail_confirmation_succeeded.html')

    def test_confirmation(self):
        """
        Tests a confirmation with a user model
        """
        obj = Confirmed()
        obj.save()

        self.assertEqual(Confirmed.objects.filter(confirmed__confirmed=True).exists(), False)

        t = 'confirm.html'
        ct = 'confirmed.html'

        generate_and_send(obj, t, ct,
                          "/success", "confirm",
                          "info@example.com", "client@example.com" )

        self.assertEqual(len(mail.outbox)>=1, True)
        confirm_url = self.r.findall(mail.outbox[-1].body)[0]

        response = self.client.get(confirm_url)

        self.assertEqual(mail.outbox[0].from_email, 'info@example.com')
        self.assertEqual(mail.outbox[0].to[0], 'client@example.com')

        self.assertEqual(MailConfirmation.objects.all()[0].confirmed, True)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(u'confirm at this url ' + confirm_url, u'confirm at this url ' + settings.SITE_SCHEME + '://' + settings.SITE_DOMAIN + reverse('mail_confirmation:url', kwargs={'confirmationid': obj.confirmed.all()[0].confirmationid}))
        self.assertEqual(Confirmed.objects.filter(confirmed__confirmed=True).exists(), True)



from celery.task.schedules import crontab
from celery.task import periodic_task

from .utils import clear_stale as clear_stale_confirmations


@periodic_task(run_every=crontab(day_of_week='1', hour='12'))
def clear_stale():
    """
    Clear stale confirmations every monday at 12pm
    """
    clear_stale_confirmations()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('mail_confirmation', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='mailconfirmation',
            name='forusers',
            field=models.ManyToManyField(to=settings.AUTH_USER_MODEL, null=True, blank=True),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import models
from django.contrib.contenttypes import generic
from django.core.mail import mail_admins

from ..models import MailConfirmation
from ..signals import confirmed as confirmed_signal

class Confirmed(models.Model):
    confirmed = generic.GenericRelation(
        MailConfirmation,
        content_type_field='toconfirm_type',
        object_id_field='toconfirm_id')

    @classmethod
    def application_confirmed_callback(cls, sender, toconfirm_type, object_id, **kwargs):
        if toconfirm_type == cls:
            url = ''.join(['http://example.com/confirmed/', str(object_id)])
            body = "check the website for that new application " + url
            mail_admins("New model application", body)


confirmed_signal.connect(Confirmed.application_confirmed_callback,
                         sender=MailConfirmation)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import uuid

from django.conf import settings
from django.db import models
from django.db import IntegrityError
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic

import logging
logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class MailConfirmation(models.Model):
    """
    Handles a confirmation token randomly generated
    """
    confirmationid = models.CharField(verbose_name=_("Confirmation id"),
                                      max_length=255,
                                      unique=True)
    confirmed = models.BooleanField(verbose_name=_("Confirmed"),
                                    default=False)
    date_created = models.DateTimeField(verbose_name=_("Date created"),
                                        auto_now_add=True)
    forusers = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True)
    confirm_success_template = models.CharField(
        verbose_name=_("Template that shows the user a success message on confirmation"),
        max_length=200,
        blank=True)
    confirm_success_url = models.CharField(
        verbose_name=_("Redirect the user to this url when the url is confirmed"),
        max_length=200,
        blank=True)
    toconfirm_type = models.ForeignKey(ContentType)
    toconfirm_id = models.PositiveIntegerField()
    toconfirm_object = generic.GenericForeignKey(
        'toconfirm_type', 'toconfirm_id')

    def save(self, *args, **kwargs):
        if self.pk is not None:
            # updating existing model
            return super(MailConfirmation, self).save(*args, **kwargs)

        retry = 10
        while retry > 0:
            retry -= 1
            self.confirmationid = uuid.uuid4().hex
            try:
                return super(MailConfirmation, self).save(*args, **kwargs)
            except IntegrityError, e:
                if e[1].find("confirmationid") == -1:
                    raise
        logger.error("Unable to generate mail_confirmation unique id, you could have won the lottery!")
        raise IntegrityError("mail_confirmation save error")

    def __str__(self):
        return self.confirmed



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding M2M table for field forusers on 'MailConfirmation'
        m2m_table_name = db.shorten_name(u'mail_confirmation_mailconfirmation_forusers')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('mailconfirmation', models.ForeignKey(orm[u'mail_confirmation.mailconfirmation'], null=False)),
            ('mailsocialuser', models.ForeignKey(orm[u'altuser.mailsocialuser'], null=False))
        ))
        db.create_unique(m2m_table_name, ['mailconfirmation_id', 'mailsocialuser_id'])


    def backwards(self, orm):
        # Removing M2M table for field forusers on 'MailConfirmation'
        db.delete_table(db.shorten_name(u'mail_confirmation_mailconfirmation_forusers'))


    models = {
        u'altuser.mailsocialuser': {
            'Meta': {'object_name': 'MailSocialUser'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'unique': 'True', 'max_length': '254'}),
            'follows': ('django.db.models.fields.related.ManyToManyField', [], {'blank': 'True', 'related_name': "u'followed'", 'null': 'True', 'symmetrical': 'False', 'to': u"orm['altuser.MailSocialUser']"}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'related_name': "u'user_set'", 'blank': 'True', 'to': u"orm['auth.Group']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'likes': ('django.db.models.fields.related.ManyToManyField', [], {'blank': 'True', 'related_name': "u'liked'", 'null': 'True', 'symmetrical': 'False', 'to': u"orm['altuser.MailSocialUser']"}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'profile_id': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'profile_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']", 'null': 'True', 'blank': 'True'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'related_name': "u'user_set'", 'blank': 'True', 'to': u"orm['auth.Permission']"})
        },
        u'auth.group': {
            'Meta': {'object_name': 'Group'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        u'auth.permission': {
            'Meta': {'ordering': "(u'content_type__app_label', u'content_type__model', u'codename')", 'unique_together': "((u'content_type', u'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'mail_confirmation.mailconfirmation': {
            'Meta': {'object_name': 'MailConfirmation'},
            'confirm_success_template': ('django.db.models.fields.CharField', [], {'max_length': '200', 'blank': 'True'}),
            'confirm_success_url': ('django.db.models.fields.CharField', [], {'max_length': '200', 'blank': 'True'}),
            'confirmationid': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'}),
            'confirmed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'date_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'forusers': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'to': u"orm['altuser.MailSocialUser']", 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'toconfirm_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'toconfirm_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"})
        }
    }

    complete_apps = ['mail_confirmation']


# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.views import redirect_to_login
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import HttpResponseRedirect
from django.views.generic.base import TemplateView
from django.shortcuts import get_object_or_404

from .models import MailConfirmation

from .signals import confirmed as confirmed_signal


class Confirm(TemplateView):

    def user_permitted(self, request, confirmation, *args, **kwargs):
        if confirmation.forusers.exists():
            if request.user not in confirmation.forusers.all():
                return False
        return True

    def dispatch(self, request, confirmationid, *args, **kwargs):
        confirmation = get_object_or_404(MailConfirmation,
                                         confirmationid=confirmationid)
        if not self.user_permitted(request, confirmation, *args, **kwargs):
            if request.user.is_anonymous():
                return redirect_to_login(request.get_full_path(),
                                         settings.LOGIN_URL,
                                         REDIRECT_FIELD_NAME)
            else:
                raise PermissionDenied

        with transaction.atomic():
            if not confirmation.confirmed:
                confirmation.confirmed = True
                confirmation.save()
                confirmed_signal.send(
                    sender=MailConfirmation,
                    toconfirm_type=confirmation.toconfirm_object.__class__,
                    object_id=confirmation.toconfirm_object.id)

            self.template_name = confirmation.confirm_success_template or confirmation.toconfirm_object._meta.app_label + '/mail_confirmation_succeeded.html'

            if confirmation.confirm_success_url:
                return HttpResponseRedirect(confirmation.confirm_success_url)

            return self.render_to_response({
                'confirmation': confirmation})



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models
from django.db.utils import OperationalError

class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'MailConfirmation'
        try:
            db.create_table(u'mail_confirmation_mailconfirmation', (
                (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('confirmationid', self.gf('django.db.models.fields.CharField')(unique=True, max_length=255)),
                ('confirmed', self.gf('django.db.models.fields.BooleanField')(default=False)),
                ('date_created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
                ('confirm_success_template', self.gf('django.db.models.fields.CharField')(max_length=200, blank=True)),
                ('confirm_success_url', self.gf('django.db.models.fields.CharField')(max_length=200, blank=True)),
                ('toconfirm_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['contenttypes.ContentType'])),
                ('toconfirm_id', self.gf('django.db.models.fields.PositiveIntegerField')()),
            ))
            db.send_create_signal(u'mail_confirmation', ['MailConfirmation'])
        except OperationalError:
            # table already exists
            pass

    def backwards(self, orm):
        # Deleting model 'MailConfirmation'
        db.delete_table(u'mail_confirmation_mailconfirmation')


    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'mail_confirmation.mailconfirmation': {
            'Meta': {'object_name': 'MailConfirmation'},
            'confirm_success_template': ('django.db.models.fields.CharField', [], {'max_length': '200', 'blank': 'True'}),
            'confirm_success_url': ('django.db.models.fields.CharField', [], {'max_length': '200', 'blank': 'True'}),
            'confirmationid': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'}),
            'confirmed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'date_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'toconfirm_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'toconfirm_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"})
        }
    }

    complete_apps = ['mail_confirmation']



# coding=utf-8
from __future__ import absolute_import, unicode_literals
from django.conf.urls import patterns, url, include

urlpatterns = patterns(
    '',
    url(r'^mail/', include('mail_confirmation.urls'
                           , namespace='mail_confirmation')),

)



import os
from setuptools import setup
from setuptools import setup, find_packages

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-mail_confirmation',
    version='1.2.2',
    packages=find_packages(),
    include_package_data=True,
    license='GPLv3 License',  # example license
    description='A general django user mail confirmation app usable with multiple models at the same time.',
    long_description=README,
    url='http://v.licheni.net/drc/django-mail_confirmation.git',
    author='Davide Riccardo Caliendo',
    author_email='davide.licheni.net',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)




"""
Copyed from https://github.com/SmileyChris/easy-thumbnails/blob/d6c5f3f8ae7a830b49ebb87fdda4b0b1d0bbf916/easy_thumbnails/migrations/__init__.py

Django migrations for mail_confirmation app
This package does not contain South migrations. South migrations can be found
in the ``south_migrations`` package.
"""
SOUTH_ERROR_MESSAGE = """\n
For South support, customize the SOUTH_MIGRATION_MODULES setting like so:
SOUTH_MIGRATION_MODULES = {
'mail_confirmation': 'mail_confirmation.south_migrations',
}
"""
# Ensure the user is not using Django 1.6 or below with South
try:
    from django.db import migrations # noqa
except ImportError:
    from django.core.exceptions import ImproperlyConfigured
    raise ImproperlyConfigured(SOUTH_ERROR_MESSAGE)



import os

SITE_ID = 1
SITE_SCHEME = 'https'
SITE_DOMAIN = 'example.com'
DATABASE_ENGINE = 'sqlite3'

ROOT_URLCONF = 'mail_confirmation.tests.urls'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

INSTALLED_APPS = [
    'django.contrib.sites',
    'django.contrib.contenttypes',
    'mail_confirmation',
    'mail_confirmation.tests',
]

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)

# This is only needed for the 1.4.X test environment
USE_TZ = True

SECRET_KEY = 'easy'
